//  Copyright (c) 2005, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <semaphore.h>
#include <string.h>

#include "Logging.h"


#ifdef ELAST
    #define CHECKPOINT ELAST
#else
    #define CHECKPOINT 125
#endif

#define RESTART (CHECKPOINT + 1)


typedef enum _CheckpointMethod {
    FastForward = 0, // Records all messages sent up to the checkpoint, which are then replayed in order in quick succession upon a restart.  Requires only that the higher layer be able to reset itself back to it's initial state.
    Rewind // Records the state of the system at the checkpoint, possibly along with any late messages that arrive during the checkpoint.  Requires higher layer to support checkpointing.
} CheckpointMethod;

#define defaultMethod Rewind
static int method = defaultMethod;

const char* nameOfCheckpointMethod(CheckpointMethod method) {
    switch (method) {
        case FastForward:
            return "FastForward"; break;
        case Rewind:
            return "Rewind"; break;
        default:
            return "Unknown";
    }
}

int fprintCheckpointMethod(FILE *output, CheckpointMethod method) {
    switch (method) {
        case FastForward:
            return fprintf(output, "FastForward"); break;
        case Rewind:
            return fprintf(output, "Rewind"); break;
        default:
            return fprintf(output, "Unknown (%d)", method);
    }
}



typedef struct _Message {
    int destination;
    int source;
    int data;
} Message;

#define MESSAGE_HEADER_SIZE (sizeof(Message)) // We presently pretend the data is part of the header, since it's fixed in size and always present.
#define MAXIMUM_MESSAGE_SIZE (sizeof(Message))

Message* allocMessage(void) {
    return (Message*)calloc(1, sizeof(Message));
}

void freeMessage(Message *message) {
    if (NULL != message) {
        free(message);
    }
}


typedef enum _Command {
    CommandData = 0,
    CommandCheckpoint,
    CommandRestart
} Command;

const char* nameOfCommand(Command command) {
    switch (command) {
        case CommandData:
            return "Data"; break;
        case CommandCheckpoint:
            return "Checkpoint"; break;
        case CommandRestart:
            return "Restart"; break;
        default:
            return "Unknown";
    }
}

int fprintCommand(FILE *output, Command command) {
    switch (command) {
        case CommandData:
            return fprintf(output, "Data"); break;
        case CommandCheckpoint:
            return fprintf(output, "Checkpoint"); break;
        case CommandRestart:
            return fprintf(output, "Restart"); break;
        default:
            return fprintf(output, "Unknown (%d)", command);
    }
}


typedef struct _LowerMessage {
    Command command;
    int sequenceNumber;
    
    Message *contents;
} LowerMessage;

#define LOWER_MESSAGE_HEADER_SIZE (sizeof(Command) + sizeof(int))
#define MAXIMUM_LOWER_MESSAGE_SIZE (sizeof(Message) + LOWER_MESSAGE_HEADER_SIZE)

LowerMessage* allocLowerMessage(void) {
    return (LowerMessage*)calloc(1, sizeof(LowerMessage));
}

void freeLowerMessage(LowerMessage *message) {
    if (NULL != message) {
        if (NULL != message->contents) {
            freeMessage(message->contents);
        }
        
        free(message);
    } else {
        PDEBUGC("Invalid parameters - message is NULL.\n");
    }
}

typedef struct _PseudoSocket {
    int pipes[2]; // For communication to/from this process.
    int outgoingSequenceNumber, // The next sequence number to use for an outgoing message.
        incomingSequenceNumber; // The next sequence number expected on an incoming message from this process.
    
    int checkpointSequenceNumber, // The sequence number at which a checkpoint command was received.  If less than 0, no such command has been received.
        restartSequenceNumber; // The sequence number at which a restart was acknowledged.  During a restart all messages are ignored until this acknowledgement is received.
} PseudoSocket;

#define RECEIVE_PIPE 0
#define SEND_PIPE 1


typedef struct option cliOption;
typedef struct sigaction signalAction;


static inline int randomInteger(int min, int max) {
    return (min + (random() % (max - min + 1)));
}


#define defaultCheckpointPosition 10
#define defaultRestartPosition 20
#define defaultNumberOfProcesses 4
#define defaultVerbosity 1

static int thisProcess = 0,
           checkpointPosition = defaultCheckpointPosition,
           restartPosition = defaultRestartPosition;
static int amCheckpointing = 0,
           haveCheckpointed = 0,
           amRestarting = 0,
           needToNotifyUpperLayerOfCheckpoint = 0;
static int verbosity = defaultVerbosity; // 0 = upper layer messages only, 1 = plus checkpoint/restart messages, 2 = plus data messages

static int numberOfBufferedMessages = 0, // The number of messages in the buffer.
           restartIteratorPosition = -1; // The position within the buffered messages when replaying via the FastForward method (not used for Rewind).
static Message **bufferedMessages = NULL; // An array of pointers to buffered Messages.

static int numberOfProcesses = defaultNumberOfProcesses;
static PseudoSocket *connections = NULL; // 'numberOfProcesses' elements
#ifdef __linux__
        static FILE *log = NULL;
#else
        static FILE *log = stdout;
#endif


#define LOG_CRITICAL 0
#define LOG_INFORMATIVE 1
#define LOG_PEDANTIC 2

#ifdef LOCKED_LOGGING
    #ifdef __linux__
        extern int flock(int fd, int operation);
        #define LOG(level, message, ...) if (verbosity >= (level)) { \
                                             flock(fileno(log), LOCK_EX); \
                                             fprintf(log, "%4d - " message, thisProcess, ## __VA_ARGS__); \
                                             fflush(log); \
                                             flock(fileno(log), LOCK_UN); \
                                         }
    #else /* __linux__ */
        static sem_t *sem;
        #define LOG(level, message, ...) if (verbosity >= (level)) { \
                                             sem_wait(sem); \
                                             fprintf(log, "%4d - " message, thisProcess, ## __VA_ARGS__); \
                                             fflush(log); \
                                             sem_post(sem); \
                                         }
    #endif /* __linux__ */
#else /* LOCKED_LOGGING */
    #define LOG(level, message, ...) if (verbosity >= (level)) { \
                                         fprintf(log, "%4d - " message, thisProcess, ## __VA_ARGS__); \
                                         fflush(log); \
                                     }
#endif /* LOCKED_LOGGING */


/*! @function sendLowerMessage
    @abstract Sends a low-level message.
    @discussion The given message needs to have all it's fields filled out appropriately.  Basic sanity checking will be performed, but nothing related to the actual data content or semantic meaning of the message.  Essentially this function just sends the message over the wire.  The bulk of it's task is just preparation, massaging of data, and error handling.
    @param message The message to send.  Should not be NULL.
    @result Returns 0 if successful, a POSIX error code otherwise (typically ENOTCONN, if the given process is no longer connected, or EIO is an error occurs writing the data [which is usually fatal]). */

int sendLowerMessage(const LowerMessage *message) {
    if ((NULL != message) && (NULL != message->contents)) {
        if ((message->contents->destination >= 0) && (message->contents->destination < numberOfProcesses)) {
            if (-1 == connections[message->contents->destination].pipes[SEND_PIPE]) {
                return ENOTCONN;
            } else {
                static char buffer[MAXIMUM_LOWER_MESSAGE_SIZE];
                int err;
                
                LOG(LOG_PEDANTIC, "Sending message to %d with sequence number %d, command %s.\n", message->contents->destination, message->sequenceNumber, nameOfCommand(message->command));
                
                memcpy(buffer, (void*)message, LOWER_MESSAGE_HEADER_SIZE);
                memcpy(buffer + LOWER_MESSAGE_HEADER_SIZE, (void*)(message->contents), MAXIMUM_MESSAGE_SIZE);
                
                err = write(connections[message->contents->destination].pipes[SEND_PIPE], (const void*)buffer, MAXIMUM_LOWER_MESSAGE_SIZE);
                
                if (-1 == err) {
                    if (EPIPE == errno) {
                        int j, allDead = 1;
                        
                        LOG(LOG_PEDANTIC, "Noticed process %d is gone.\n", message->contents->destination);
                        close(connections[message->contents->destination].pipes[SEND_PIPE]);
                        connections[message->contents->destination].pipes[SEND_PIPE] = -1;
                        
                        // This is probably unnecessary - on the next call to select we'll notice our read pipe has closed anyway.
                        
                        for (j = 0; (1 == allDead) && (j < numberOfProcesses); ++j) {
                            if (-1 != connections[j].pipes[SEND_PIPE]) {
                                allDead = 0;
                            }
                        }
                        
                        if (1 == allDead) {
                            LOG(LOG_PEDANTIC, "All other processes appear to be gone.\n");
                            return ENETDOWN;
                        }
                        
                        return ENOTCONN;
                    } else {
                        PERRC("Unable to write data to process #%d, error #%d (%s).\n", message->contents->destination, errno, strerror(errno));
                        return EIO;
                    }
                } else if (MAXIMUM_LOWER_MESSAGE_SIZE != err) {
                    PERRC("Unable to write the entire message to process #%d (wanted to write %d bytes, but wrote %d).\n", message->contents->destination, MAXIMUM_LOWER_MESSAGE_SIZE, err);
                    return EIO;
                } else {
                    return 0;
                }
            }
        } else {
            PDEBUGC("Message destination is invalid (%d, with %d processes).\n", message->contents->destination, numberOfProcesses);
            return EFAULT;
        }
    } else {
        PDEBUGC("Invalid parameters (message = %p).\n", message);
        return EINVAL;
    }
}

/*! @function sendMessage
    @abstract Sends a message.
    @param message The message to send.  Should not be NULL.
    @param command The command specifying the type of message being sent.
    @result Returns 0 if successful, a POSIX error code otherwise (typically ENOTCONN if the desired process is no longer connected, EINVAL if invalid parameters are supplied, or EIO if an error occurs trying to send the data [which is usually fatal]). */

int sendMessage(Message *message, Command command) {
    if (NULL != message) {
        if (thisProcess != message->destination) {
            if ((0 <= message->destination) && (numberOfProcesses > message->destination)) {
                LowerMessage messageToSend;
                int err;
                
                messageToSend.command = command;
                messageToSend.sequenceNumber = connections[message->destination].outgoingSequenceNumber++;
                messageToSend.contents = message;
                
                //LOG(LOG_PEDANTIC, "Sending message to %d.\n", message->destination);
                
                err = sendLowerMessage(&messageToSend);
                
                return err;
            } else {
                PERRC("Invalid destination, %d (number of processes = %d).\n", message->destination, numberOfProcesses);
                return ENOENT;
            }
        } else {
            PERRC("Tried to send a message to myself... whoops.\n");
            return EADDRNOTAVAIL;
        }
    } else {
        PERRC("Invalid parameter (message = %p).\n", message);
        return EINVAL;
    }
}

/*! @function sendData
    @abstract Sends some data.
    @discussion Intended to be used by the application layer itself, this function's inputs are trivial - a destination, a command (a.k.a. message type) and some data (presently just an integer, but in future may be extended to variable length binary data as required).
    @param destination The destination process ID.  Should be greater than or equal to zero, and less than 'numberOfProcesses', nor correspond to the sending process.
    @param data The data to send.  This is compulsory.  If in doubt, use an appropriate pre-defined "NULL" value - e.g. 0 or -1.
    @result Returns 0 if successful, a POSIX error code otherwise (typically ENOTCONN if the desired process is no longer connected, EINVAL if invalid parameters are supplied, or EIO if an error occurs trying to send the data [which is usually fatal]). */

int sendData(int destination, int data) {
    if ((thisProcess != destination) && (0 <= destination) && (destination < numberOfProcesses) && (destination != thisProcess)) {
        Message contents;
        
        contents.destination = destination;
        contents.source = thisProcess;
        contents.data = data;
        
        //LOG(LOG_PEDANTIC, "Sending data message to %d.\n", destination);
        
        return sendMessage(&contents, CommandData);
    } else {
        PDEBUGC("Invalid parameters (destination = %d [thisProcess = %d], data = %d).\n", destination, thisProcess, data);
        return EINVAL;
    }
}


int startCheckpoint(void) {
    if ((0 == haveCheckpointed) && (0 == amCheckpointing)) {
        int err = 0;
        
        if (0 == amRestarting) {
            int i;
            Message checkpointMessage;
            
            amCheckpointing = 1;
            needToNotifyUpperLayerOfCheckpoint = 1;
            
            checkpointMessage.source = thisProcess;
            checkpointMessage.data = 0;
            
            for (i = 0; (0 == err) && (i < numberOfProcesses); ++i) {
                if (i != thisProcess) {
                    checkpointMessage.destination = i;
                    
                    err = sendMessage(&checkpointMessage, CommandCheckpoint);
                    
                    if ((0 != err) && (ENOTCONN != err)) {
                        PERRC("Unable to send checkpoint message to %d, error #%d (%s).\n", i, err, strerror(err));
                    } else {
                        err = 0;
                    }
                }
            }
        } else {
            PERRC("Tried to restart while a checkpoint was in progress.\n");
            err = EAGAIN;
        }
        
        return err;
    } else {
        return 0;
    }
}

int startRestart(void) {
    if (0 == amRestarting) {
        int err = 0;
        
        if (0 == amCheckpointing) {
            if (0 != haveCheckpointed) {
                int i;
                Message restartMessage;
                
                amRestarting = 1;
                
                restartMessage.source = thisProcess;
                restartMessage.data = 0;
                
                for (i = 0; (0 == err) && (i < numberOfProcesses); ++i) {
                    if (i != thisProcess) {
                        restartMessage.destination = i;
                        
                        err = sendMessage(&restartMessage, CommandRestart);
                        
                        if ((0 != err) && (ENOTCONN != err)) {
                            PERRC("Unable to send restart message to %d, error #%d (%s).\n", i, err, strerror(err));
                        } else {
                            err = 0;
                        }
                    }
                }
            } else {
                PERRC("Tried to restart without having first performed a checkpoint.\n");
                err = EAGAIN;
            }
        } else {
            PERRC("Tried to restart while a checkpoint was occuring.\n");
            err = EAGAIN;
        }
        
        return err;
    } else {
        return 0;
    }
}

/*! @function waitforLowerMessageWithTimeout
    @abstract Waits for a lower message to arrive, returning nothing if a given number of microseconds elapse without anything arriving.
    @discussion Pretty straight foward.
    @param message On output, is set to the first message to arrive.  If set to NULL, consult the return value - if it is 0, the timeout was encountered before a message arrived, if negative an error occured.  Should not be NULL.  The caller is responsible for freeing the returned message (if any) using the freeMessage() function.
    @param timeoutInMicroseconds The maximum number of microseconds to wait for a message to arrive before returning.  If 0 the function returns immediately, checking for new messages only in local buffers (i.e. already received).
    @result Returns the number of microseconds remaining of the timeout (i.e. positive) or, if an error occurs, a POSIX error code (i.e. negative). */

int waitforLowerMessageWithTimeout(LowerMessage **message, int timeoutInMicroseconds) {
    if ((NULL != message) || (0 > timeoutInMicroseconds)) {
        fd_set selectors;
        struct timeval startTime, endTime, timeout;
        char buffer[MAXIMUM_LOWER_MESSAGE_SIZE];
        int err;
        unsigned int timePassed;
        
        *message = NULL;
        
        while (NULL == *message) {
            FD_ZERO(&selectors);
            FD_SET(connections[thisProcess].pipes[RECEIVE_PIPE], &selectors);
            
            timeout.tv_sec = timeoutInMicroseconds / 1000000;
            timeout.tv_usec = timeoutInMicroseconds % 1000000;
            
            err = gettimeofday(&startTime, NULL);
            
            if (0 != err) {
                PERRC("Unable to get the time of day, error #%d (%s).\n", errno, strerror(errno));
                return -errno;
            }
            
            err = select(connections[thisProcess].pipes[RECEIVE_PIPE] + 1, &selectors, NULL, NULL, &timeout);
            
            if (-1 == err) {
                PERRC("Unable to select pipe, error #%d (%s).\n", errno, strerror(errno));
                return -errno;
            } else {
                err = gettimeofday(&endTime, NULL);
                
                if (0 != err) {
                    PERRC("Unable to get the time of day, error #%d (%s).\n", errno, strerror(errno));
                    return -errno;
                }
                
                timePassed = (((endTime.tv_sec - startTime.tv_sec) * 1000000) + ((long)endTime.tv_usec - (long)startTime.tv_usec));
                
                if (timePassed > timeoutInMicroseconds) {
                    timeoutInMicroseconds = 0;
                } else {
                    timeoutInMicroseconds -= timePassed;
                }
                
                if (FD_ISSET(connections[thisProcess].pipes[RECEIVE_PIPE], &selectors)) { // We've got a message
                    err = read(connections[thisProcess].pipes[RECEIVE_PIPE], (void*)buffer, MAXIMUM_LOWER_MESSAGE_SIZE);
                    
                    if (-1 == err) {
                        PERRC("Unable to read from my pipe, despite select() saying it's got data - error #%d (%s).\n", errno, strerror(errno));
                        return -errno;
                    } else if (0 == err) {
                        LOG(LOG_PEDANTIC, "All other processes appear to have finished.\n");
                        return -ENETDOWN;
                    } else if (MAXIMUM_LOWER_MESSAGE_SIZE != err) {
                        PERRC("Unable to read an entire message from the pipe (expected %d bytes, got %d).\n", MAXIMUM_LOWER_MESSAGE_SIZE, err);
                        return -EIO;
                    } else {
                        *message = allocLowerMessage();
                        
                        if (NULL == *message) {
                            PDEBUGC("Unable to allocate new LowerMessage.\n");
                            return -ENOMEM;
                        }
                        
                        (*message)->contents = allocMessage();
                        
                        if (NULL == (*message)->contents) {
                            PDEBUGC("Unable to allocate new Message.\n");
                            return -ENOMEM;
                        }
                        
                        memcpy((void*)(*message), buffer, LOWER_MESSAGE_HEADER_SIZE);
                        memcpy((void*)(*message)->contents, buffer + LOWER_MESSAGE_HEADER_SIZE, MAXIMUM_MESSAGE_SIZE);
                        
                        LOG(LOG_PEDANTIC, "Received message destined for %d, from %d, with data %d, sequence number %d, and command %s.\n", (*message)->contents->destination, (*message)->contents->source, (*message)->contents->data, (*message)->sequenceNumber, nameOfCommand((*message)->command));
                        
                        if ((0 <= (*message)->contents->source) && ((*message)->contents->source < numberOfProcesses)) {
                            if ((*message)->contents->destination != thisProcess) {
                                PDEBUGC("The destination process ID %d of the received message is not ourselves (%d).\n", (*message)->contents->destination, thisProcess);
                                return -EIO;
                            } else {
                                if ((*message)->sequenceNumber < connections[(*message)->contents->source].incomingSequenceNumber) {
                                    PDEBUGC("The sequence number of %d is not the %d expected.  Ignoring; assuming duplicate of old message.\n", (*message)->sequenceNumber, connections[(*message)->contents->source].incomingSequenceNumber);
                                } else if ((*message)->sequenceNumber < connections[(*message)->contents->source].incomingSequenceNumber) {
                                    PDEBUGC("The sequence number of %d is not the %d expected.  Buffering of out-of-sequence messages not yet supported; barfing.\n", (*message)->sequenceNumber, connections[(*message)->contents->source].incomingSequenceNumber);
                                    return -EIO;
                                }
                            }
                        } else {
                            PDEBUGC("The source process ID %d of the received message is not valid (given there are %d processes).\n", (*message)->contents->source, numberOfProcesses);
                            return -EIO;
                        }
                    }
                } else if (0 == timeoutInMicroseconds) {
                    break;
                }
            }
        }
        
        return timeoutInMicroseconds;
    } else {
        PDEBUGC("Invalid parameters (message = %p, timeoutInMicroseconds = %u).\n", message, timeoutInMicroseconds);
        return -EINVAL;
    }
}

/*! @function waitForMessageWithTimeout
    @abstract Waits for a message to arrive, returning nothing if a given number of microseconds elapse without a message arriving.
    @discussion Pretty straight foward.
    @param message On output, is set to the first message to arrive.  If set to NULL, consult the return value - if it is 0, the timeout was encountered before a message arrived, if negative an error occured.  Should not be NULL.  The caller is responsible for freeing the returned message (if any) using the freeMessage() function.
    @param timeoutInMicroseconds The maximum number of microseconds to wait for a message to arrive before returning.  If 0 the function returns immediately, checking for new messages only in local buffers.
    @result Returns the number of microseconds remaining of the timeout (i.e. positive) or, if an error occurs, a POSIX error code (i.e. negative). */

int waitForMessageWithTimeout(Message **message, int timeoutInMicroseconds) {
    if ((NULL != message) && (0 <= timeoutInMicroseconds)) {
        LowerMessage *lowerMessage;
        int timeoutRemaining = timeoutInMicroseconds;
        int i, haveFinishedCheckpointing, haveFinishedRestarting;
        
        if ((Rewind == method) && (0 != needToNotifyUpperLayerOfCheckpoint)) {
            timeoutRemaining = -CHECKPOINT;
            needToNotifyUpperLayerOfCheckpoint = 0;
        } else if ((0 != amRestarting) && ((0 > restartIteratorPosition) || ((Rewind == method) && (0 < numberOfBufferedMessages)) || ((FastForward == method) && (restartIteratorPosition < numberOfBufferedMessages)))) {
            if (0 > restartIteratorPosition) {
                restartIteratorPosition = 0;
                timeoutRemaining = -RESTART;
            } else {
                if (FastForward == method) {
                    *message = allocMessage();
                    
                    if (NULL == *message) {
                        PERRC("Unable to allocate memory for copy of buffered message.\n");
                        timeoutRemaining = -ENOMEM;
                    } else {
                        **message = *(bufferedMessages[restartIteratorPosition]);
                        ++restartIteratorPosition;
                    }
                } else { // Rewind
                    *message = bufferedMessages[0];

                    --numberOfBufferedMessages;
                    
                    if (0 == numberOfBufferedMessages) {
                        free(bufferedMessages);
                        bufferedMessages = NULL;
                    } else {
                        memmove(bufferedMessages, bufferedMessages + 1, sizeof(Message*) * numberOfBufferedMessages);
                    }
                }
                
                LOG(LOG_INFORMATIVE, "Replaying message from %d with data %d.\n", (*message)->source, (*message)->data);
            }
        } else {
            *message = NULL;
            
            do {
                timeoutRemaining = waitforLowerMessageWithTimeout(&lowerMessage, timeoutRemaining);
                
                if (0 <= timeoutRemaining) {
                    if (NULL != lowerMessage) {
                        switch (lowerMessage->command) {
                            case CommandData:
                                if ((0 == amRestarting) || ((0 <= connections[lowerMessage->contents->source].restartSequenceNumber) && (lowerMessage->sequenceNumber > connections[lowerMessage->contents->source].restartSequenceNumber))) {
                                    if (((FastForward == method) || ((Rewind == method) && (0 != amCheckpointing))) && (0 > connections[lowerMessage->contents->source].checkpointSequenceNumber)) {
                                        LOG(LOG_INFORMATIVE, "Received data from %d - buffering for checkpointing purposes.\n", lowerMessage->contents->source);
                                        
                                        ++numberOfBufferedMessages;
                                        bufferedMessages = realloc(bufferedMessages, sizeof(Message*) * numberOfBufferedMessages);
                                        
                                        if (NULL == bufferedMessages) {
                                            PERRC("Unable to add message to buffer (%d items already in buffer).\n", numberOfBufferedMessages - 1);
                                            return -ENOMEM;
                                        }
                                        
                                        bufferedMessages[numberOfBufferedMessages - 1] = allocMessage();
                                        
                                        if (NULL == bufferedMessages[numberOfBufferedMessages - 1]) {
                                            PERRC("Unable to allocate memory for extra message to buffer.\n");
                                            return -ENOMEM;
                                        }
                                        
                                        *(bufferedMessages[numberOfBufferedMessages - 1]) = *(lowerMessage->contents);
                                    }
                                    
                                    *message = lowerMessage->contents;
                                    lowerMessage->contents = NULL;
                                } else {
                                    LOG(LOG_CRITICAL, "Received old data (%d) from %d during a restart - ignoring.\n", lowerMessage->contents->data, lowerMessage->contents->source);
                                }
                                
                                break;
                            case CommandCheckpoint:
                                if (0 != haveCheckpointed) {
                                    PERRC("Received checkpoint message from %d, but a checkpoint has already been performed; ignoring.\n", lowerMessage->contents->source);
                                } else {
                                    LOG(LOG_INFORMATIVE, "Received checkpoint token from %d.\n", lowerMessage->contents->source);
                                    
                                    if (0 == amCheckpointing) {
                                        int err;
                                        
                                        LOG(LOG_INFORMATIVE, "Beginning checkpoint...\n");
                                        
                                        if (Rewind == method) {
                                            timeoutRemaining = -CHECKPOINT;
                                        }

                                        err = startCheckpoint();
                                        
                                        if (0 != err) {
                                            PERRC("Unable to start checkpointing, error #%d (%s).\n", err, strerror(err));
                                            return -err;
                                        } else {
                                            needToNotifyUpperLayerOfCheckpoint = 0;
                                        }
                                    }
                                    
                                    if (0 <= connections[lowerMessage->contents->source].checkpointSequenceNumber) {
                                        PERRC("Already received checkpoint message from %d.\n", lowerMessage->contents->source);
                                        return -EFAULT;
                                    }
                                    
                                    connections[lowerMessage->contents->source].checkpointSequenceNumber = lowerMessage->sequenceNumber;
                                    connections[lowerMessage->contents->source].restartSequenceNumber = -1;

                                    haveFinishedCheckpointing = 1;
                                    
                                    for (i = 0; (1 == haveFinishedCheckpointing) && (i < numberOfProcesses); ++i) {
                                        if ((thisProcess != i) && (0 > connections[i].checkpointSequenceNumber) && (0 <= connections[i].pipes[SEND_PIPE])) {
                                            haveFinishedCheckpointing = 0;
                                        }
                                    }
                                    
                                    if (1 == haveFinishedCheckpointing) {
                                        LOG(LOG_INFORMATIVE, "Checkpoint complete.\n");
                                        amCheckpointing = 0;
                                        haveCheckpointed = 1;
                                        
                                        restartIteratorPosition = -1;
                                    } else {
                                        //LOG(LOG_PEDANTIC, "Checkpoint still incomplete.\n");
                                    }
                                }
                                
                                break;
                            case CommandRestart:
                                if (1 == amCheckpointing) {
                                    PERRC("Received restart from %d during a checkpoint.\n", lowerMessage->contents->source);
                                    return -EFAULT;
                                } else if (0 > connections[lowerMessage->contents->source].checkpointSequenceNumber) {
                                    PERRC("Received restart from %d without having previously performed a checkpoint.\n", lowerMessage->contents->source);
                                    return -EFAULT;
                                } else {
                                    LOG(LOG_INFORMATIVE, "Received restart from %d.\n", lowerMessage->contents->source);
                                    
                                    if (0 == amRestarting) {
                                        int err;

                                        LOG(LOG_INFORMATIVE, "Beginning restart...\n");
                                        
                                        timeoutRemaining = -RESTART;
                                        restartIteratorPosition = 0;

                                        err = startRestart();
                                        
                                        if (0 != err) {
                                            PERRC("Unable to start restart (as prompted by %d), error #%d (%s).\n", lowerMessage->contents->source, err, strerror(err));
                                            timeoutRemaining = -err;
                                        }
                                    }
                                    
                                    if (0 <= connections[lowerMessage->contents->source].restartSequenceNumber) {
                                        PERRC("Already received restart message from %d.\n", lowerMessage->contents->source);
                                        return -EFAULT;
                                    }
                                    
                                    connections[lowerMessage->contents->source].restartSequenceNumber = lowerMessage->sequenceNumber;
                                    connections[lowerMessage->contents->source].checkpointSequenceNumber = -1;

                                    haveFinishedRestarting = 1;
                                    
                                    for (i = 0; (1 == haveFinishedRestarting) && (i < numberOfProcesses); ++i) {
                                        if ((thisProcess != i) && (0 > connections[i].restartSequenceNumber) && (0 <= connections[i].pipes[SEND_PIPE])) {
                                            haveFinishedRestarting = 0;
                                        }
                                    }
                                    
                                    if (1 == haveFinishedRestarting) {
                                        LOG(LOG_INFORMATIVE, "Restart complete.\n");
                                        amRestarting = 0;
                                        
                                        haveCheckpointed = 0;
                                    } else {
                                        //LOG(LOG_PEDANTIC, "Restart still incomplete.\n");
                                    }
                                }
                                
                                break;
                            default:
                                PERRC("Received unknown command type in message, %d.\n", lowerMessage->command);
                        }
                        
                        freeLowerMessage(lowerMessage);
                    } // Else just timed out; fall through
                } else {
                    if (-ENETDOWN != timeoutRemaining) {
                        PERRC("Unable to retrieve lower message, error #%d (%s).\n", -timeoutRemaining, strerror(-timeoutRemaining));
                    }
                }
            } while ((NULL == *message) && (0 < timeoutRemaining));
        }
        
        return timeoutRemaining;
    } else {
        PDEBUGC("Invalid parameters (message = %p, timeoutInMicroseconds = %u).\n", message, timeoutInMicroseconds);
        return -EINVAL;
    }
}

int main(int argc, char* const argv[]) {
    const int defaultNumberOfMessages = 30,
              defaultMinimumWaitTime = 0,
              defaultMaximumWaitTime = 1000000;
    int numberOfMessages = defaultNumberOfMessages,
        minimumWaitTime = defaultMinimumWaitTime,
        maximumWaitTime = defaultMaximumWaitTime;
    cliOption longopts[] = { { "checkpoint", required_argument, NULL, 'c' },
                             { "logfile", required_argument, NULL, 'f' },
                             { "help", no_argument, NULL, 'h' },
                             { "method", required_argument, NULL, 'M' },
                             { "messages", required_argument, NULL, 'm' },
                             { "processes", required_argument, NULL, 'p' },
                             { "restart", required_argument, NULL, 'r' },
                             { "verbosity", required_argument, NULL, 'v' },
                             { "minwaittime", required_argument, NULL, '-' },
                             { "maxwaittime", required_argument, NULL, '+' },
                             { NULL, 0, NULL, 0 } };
    signalAction signal;
    int opt, i, err, work = 0, savedWork;
    int timeout;
    Message *message;
    int destination;

#ifdef __linux__
    log = stdout;
#endif

#if defined(LOCKED_LOGGING) && defined(__linux__)
    sem = sem_open("checkpoint", O_CREAT | O_EXCL);
    
    if (NULL == sem) {
        PERRC("Unable to initialise semaphore, error #%d (%s).\n", errno, strerror(errno));
        exit(errno);
    } else {
        sem_post(sem);
    }
#endif
    
    signal.sa_flags = 0;
    signal.sa_handler = SIG_IGN;
    sigemptyset(&(signal.sa_mask));
    
    err = sigaction(SIGPIPE, &signal, NULL);
    
    if (0 != err) {
        PERRC("Unable to install signal handler, error #%d (%s).\n", errno, strerror(errno));
        exit(errno);
    }
    
    // Parse an CLI arguments
    
    while (-1 != (opt = getopt_long(argc, argv, "c:f:hM:m:p:r:v:", longopts, NULL))) {
        switch (opt) {
            case 'c': // Checkpoint
                {
                    char *end;
                
                    checkpointPosition = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (0 >= checkpointPosition)) {
                        PERRC("Invalid parameter to the --checkpoint/-c argument, \"%s\".  Expected a positive integer greater than 0 (and less than the restart value).\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case 'f': // Log file
                log = fopen(optarg, "w");
                
                if (NULL == log) {
                    PERRC("Unable to open logfile \"%s\"; error #%d (%s).\n", optarg, errno, strerror(errno));
                    exit(errno);
                }
                
                break;
            case 'h': // Help/usage info
                POUTC("Usage: %s [OPTIONS]\n\n..where OPTIONS may be any of:\n\n", argv[0]);
                POUTC("\t--checkpoint/-c\tTakes an integer parameter which is the message sequence number at which to initiate a checkpoint.  Must be less than the restart position, and greater than zero.  Defaults to %d.\n", defaultCheckpointPosition);
                POUTC("\t--logfile/-f\tTakes a path as it's parameter, indicating the file to log events to.  Defaults to standard output.\n");
                POUTC("\t--help/-h\tPrints this usage information.\n");
                POUTC("\t--method/-M\tTakes a string argument specifying the checkpointing method to be used, one of either \"rewind\" or \"fastforward\".  Defaults to %s.\n", nameOfCheckpointMethod(defaultMethod));
                POUTC("\t--messages/-m\tTakes an integer parameter which is the number of messages each process should send.  Defaults to %d.\n", defaultNumberOfMessages);
                POUTC("\t--maxwaittime\tTakes an integer parameter which is the maximum number of microseconds to wait between messages (per individual client).  Defaults to %d.\n", defaultMaximumWaitTime);
                POUTC("\t--minwaittime\tTake an integer parameter which is the minimum number of microseconds to wait between messages (per individual client).  Defaults to %d.\n", defaultMinimumWaitTime);
                POUTC("\t--processes/-p\tTakes an integer parameter which is the number of processes to spawn for messaging.  Defaults to %d.\n", defaultNumberOfProcesses);
                POUTC("\t--restart/-r\tTakes an integer parameter which is the message sequence number at which to initiate a restart.  Must be greater than the checkpoint position.  Defaults to %d.\n", defaultRestartPosition);
                POUTC("\t--verbosity/-v\tTakes an integer parameter which sets the verbosity of the program's output, ranging from 0 (only critical messages) to 2 (very pedantic).  A negative value indicates only the first process (#0) should log any output.  Defaults to %d.\n", defaultVerbosity);
                
                exit(0);
                
                break;
            case 'M': // Method [used for checkpointing]
                if (0 == strcasecmp(optarg, "rewind")) {
                    method = Rewind;
                } else if (0 == strcasecmp(optarg, "fastforward")) {
                    method = FastForward;
                } else {
                    PERRC("Unknown checkpoint method, \"%s\".  See \"%s -h\" for usage information.\n", optarg, argv[0]);
                    exit(EINVAL);
                }
                
                break;
            case 'm': // Number of messages
                {
                    char *end;
                
                    numberOfMessages = strtol(optarg, &end, 0);
                    
                    if ((NULL == end) || (0 != *end) || (0 >= numberOfMessages)) {
                        PERRC("Invalid parameter to the --messages/-m argument, \"%s\".  Expected a positive integer greater than 0.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case 'p': // Number of processes
                {
                    char *end;
                
                    numberOfProcesses = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (1 >= numberOfProcesses)) {
                        PERRC("Invalid parameter to the --processes/-p argument, \"%s\".  Expected a positive integer greater than 1.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case 'r': // Restart
                {
                    char *end;
                
                    restartPosition = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (0 >= restartPosition)) {
                        PERRC("Invalid parameter to the --restart/-r argument, \"%s\".  Expected a positive integer greater than the checkpoint value.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case 'v': // Verbosity
            {
                char *end;
                
                verbosity = strtol(optarg, &end, 0);
                
                if ((NULL == end) || (0 != *end) || (-2 > verbosity) || (2 < verbosity)) {
                    PERRC("Invalid parameter to the --verbosity/-v argument, \"%s\".  Expected an integer between -2 and 2 (both inclusive).\n", optarg);
                    exit(EINVAL);
                }
            }
                
                break;
            case '-': // Minimum wait time
                {
                    char *end;
                
                    minimumWaitTime = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (0 > minimumWaitTime)) {
                        PERRC("Invalid parameter to the --minwaittime argument, \"%s\".  Expected a positive integer greater than or equal to 0.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            case '+': // Maximum wait time
                {
                    char *end;
                
                    maximumWaitTime = strtol(optarg, &end, 0);
                
                    if ((NULL == end) || (0 != *end) || (0 >= maximumWaitTime)) {
                        PERRC("Invalid parameter to the --maxwaittime argument, \"%s\".  Expected a positive integer greater than 0.\n", optarg);
                        exit(EINVAL);
                    }
                }
                
                break;
            default:
                PERRC("Unknown option.  See \"%s -h\" for usage information.\n", argv[0]);
                exit(EINVAL);
        }
    }
    
    if (minimumWaitTime > maximumWaitTime) {
        PERRC("The maximum wait time must be greater than the minimum wait time (%d vs %d, respectively).\n", maximumWaitTime, minimumWaitTime);
        exit(EINVAL);
    }
    
    if (checkpointPosition >= restartPosition) {
        PERRC("The checkpoint position (%d) must be before the restart position (%d).\n", checkpointPosition, restartPosition);
        exit(EINVAL);
    }
    
    // Allocate the PseudoSockets we'll use for communication.
    
    connections = calloc(1, sizeof(PseudoSocket) * numberOfProcesses);
    
    if (NULL == connections) {
        PERRC("Unable to allocate memory (%d bytes) for connections.\n", sizeof(PseudoSocket) * numberOfProcesses);
        exit(ENOMEM);
    }
        
    for (i = 0; i < numberOfProcesses; ++i) {
        if (0 != pipe(connections[i].pipes)) {
            PERRC("Unable to create pipe pair #%d, error #%d (%s).\n", i, errno, strerror(errno));
            exit(errno);
        }
        
        connections[i].incomingSequenceNumber = connections[i].outgoingSequenceNumber = 0;
        connections[i].checkpointSequenceNumber = connections[i].restartSequenceNumber = -1;
    }
    
    // Fork off our numerous processes
    
    for (i = 1; i < numberOfProcesses; ++i) {
        PDEBUGC("Forking #%d...\n", i);
        
        err = fork();
        
        if (-1 == err) {
            PERRC("Unable to fork process #%d, error #%d (%s).\n", i, errno, strerror(errno));
            exit(errno);
        } else if (0 == err) { // We're the child, so exit this loop, after setting or process number.
            thisProcess = i;
            break;
        } // Else we're the parent; continue on.
    }
    
    if ((0 > verbosity) && (0 == thisProcess)) {
        verbosity = -verbosity;
    }
    
    srandom(time(NULL));

    // Now go through and close the read side of every pipe except our own, for which we close the write side.  This is not strictly necessary, but is a good idea as a way of conserving resources, and ensuring any mistakes in the code won't be able to send or receive messages from semantically invalid pipes.
    
    for (i = 0; i < numberOfProcesses; ++i) {
        if (thisProcess != i) {
            close(connections[i].pipes[RECEIVE_PIPE]);
            connections[i].pipes[RECEIVE_PIPE] = -1;
        } else {
            close(connections[i].pipes[SEND_PIPE]);
            connections[i].pipes[SEND_PIPE] = -1;
        }
    }
    
    // Now we can send our messages.
    
    err = 0;
    
    for (i = 0; (0 == err) && ((i < numberOfMessages) || (0 != haveCheckpointed)); ++i) {
        timeout = randomInteger(minimumWaitTime, maximumWaitTime);
        
        ++work;
        
        do {
            timeout = waitForMessageWithTimeout(&message, timeout);
            
            if (0 > timeout) {
                switch (timeout) {
                    case -ENETDOWN:
                        LOG(LOG_CRITICAL, "Exiting with %d messages unsent.\n", numberOfMessages - i);
                        exit(0);
                        break;
                    case -CHECKPOINT:
                        LOG(LOG_CRITICAL, "Notified of checkpoint - saving state (%d).\n", work);
                        
                        savedWork = work;
                        
                        break;
                    case -RESTART:
                        if (FastForward == method) {
                            LOG(LOG_CRITICAL, "Notified of restart - reseting.\n");
                            
                            work = 0;
                        } else {
                            LOG(LOG_CRITICAL, "Notified of restart - restoring state (%d).\n", savedWork);
                            
                            work = savedWork;
                        }
                        
                        break;
                    default:
                        PERRC("Error #%d (%s) occured while waiting for a message.\n", -timeout, strerror(-timeout));
                        exit(-timeout);
                }
            } else if (0 < timeout) {
                if (NULL != message) {
                    LOG(LOG_CRITICAL, "Got message from %d with data %d.\n", message->source, message->data);
                    
                    work += message->data;
                    
                    freeMessage(message);
                } else {
                    PERRC("Expected a message, and seemed to get one, but it's gone missing... d'oh.\n", thisProcess);
                    exit(EFAULT);
                }
            }
        } while (0 < timeout);
        
        if ((0 == amRestarting) && (i < numberOfMessages)) {
            do {
                do {
                    destination = randomInteger(0, numberOfProcesses - 1);
                } while (thisProcess == destination);
                
                err = sendData(destination, work);
            } while ((ENOTCONN == err) || (EINVAL == err));
            
            if (ENETDOWN == err) {
                LOG(LOG_CRITICAL, "All other processes have gone offline.  Exiting with %d message%s unsent.\n", numberOfMessages - i, ((numberOfMessages - i) != 1) ? "s" : "");
                exit(0);
            } else if (0 != err) {
                PERRC("Received error #%d (%s) from sendMessage - unable to handle.\n", err, strerror(err));
                exit(err);
            } else if ((0 == amCheckpointing) && (0 == haveCheckpointed) && (checkpointPosition == (i + 1))) {
                LOG(LOG_CRITICAL, "Initiating checkpoint, having sent %d messages.\n", i + 1);
                
                err = startCheckpoint();
                
                if (0 != err) {
                    PERRC("Unable to start checkpoint, error #%d (%s).\n", err, strerror(err));
                }
            } else if ((0 != haveCheckpointed) && (0 == amRestarting) && (restartPosition == (i + 1))) {
                LOG(LOG_CRITICAL, "Restarting, having sent %d messages.\n", i + 1);
                
                err = startRestart();
                
                if (0 != err) {
                    PERRC("Unable to start restart, error #%d (%s).\n", err, strerror(err));
                }
            }
        }
    }
    
    if (0 == err) {
        LOG(LOG_CRITICAL, "All messages sent, exiting.\n");
    }
    
    return err;
}
